open Ast

(* The key thing for the semantic checked ast *)
type sexpr = typ * sx
and sx =
    SLit of int
  | SStrlit of string
  | SLintlit of string
  | SPtlit of sexpr * sexpr * sexpr
  | SPolylit of sexpr * sexpr * sexpr
  | SAccess of string * int
  | SId of string
  | SBinop of sexpr * operator * sexpr
  | SRelop of sexpr * operator * sexpr
  | SUnop of uoperator * sexpr
  | STrnop of sexpr * toperator * sexpr * toperator * sexpr
  | SAssign of string * sexpr
  | SCall of string * sexpr list
  | SNoexpr

type sstmt =
    SBlock of sstmt list
  | SExpr of sexpr
  | SReturn of sexpr
  | SIf of sexpr * sstmt * sstmt (*just for pretty print for now*)
  | SFor of sexpr * sexpr * sexpr * sstmt
  | SWhile of sexpr * sstmt

type sfunc_decl = {
    styp : typ;
    sname : string;
    sparams: bind list;
    slocals : bind list;
    sbody : sstmt list;
}

type sprogram = bind list * sfunc_decl list

(* Pretty-printing functions *)

let rec string_of_sexpr (t, e) =
  "(" ^ string_of_typ t ^ " : " ^ (match e with
    SStrlit(l) -> "\"" ^ l ^ "\""
  | SLintlit(l) -> l
  | SPtlit(i, j, p) -> "[" ^ string_of_sexpr i ^ "," ^ string_of_sexpr j ^ "] & " ^ string_of_sexpr p
  | SPolylit(i, j, m) -> "[(" ^ string_of_sexpr i ^ "," ^ string_of_sexpr j^ ") : " ^string_of_sexpr m ^ "]"
  | SAccess(s, i) ->s ^ "." ^ string_of_int i
  | SLit(l) -> string_of_int l
  | SId(s) -> s
  | SBinop(e1, o, e2) ->
          string_of_sexpr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr e2
  | SRelop(e1, o, e2) ->
          string_of_sexpr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr e2
  | SUnop(o, e) -> string_of_uop o ^ string_of_sexpr e
  | STrnop(e1, o1, e2, o2, e3) ->
          string_of_sexpr e1 ^ " " ^ string_of_top o1 ^ " " ^ string_of_sexpr e2 ^ " " ^
            string_of_top o2 ^ " " ^ string_of_sexpr e3
  | SCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SAssign(v, e) -> v ^ " = " ^ string_of_sexpr e
  | SNoexpr -> ""
				  ) ^ ")"

let rec string_of_sstmt = function
    SBlock(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SExpr(expr) -> string_of_sexpr expr ^ ";\n";
  | SReturn(expr) -> "return " ^ string_of_sexpr expr ^ ";\n";
  | SIf(e, s, SBlock([])) -> 
      "if (" ^ string_of_sexpr e ^ ")\n" ^ string_of_sstmt s
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr e ^ ")\n" ^
      string_of_sstmt s1 ^ "else\n" ^ string_of_sstmt s2
  | SFor(e1, e2, e3, s) ->
      "for (" ^ string_of_sexpr e1  ^ " ; " ^ string_of_sexpr e2 ^ " ; " ^
      string_of_sexpr e3  ^ ") " ^ string_of_sstmt s
  | SWhile(e, s) -> "while (" ^ string_of_sexpr e ^ ") " ^ string_of_sstmt s

let string_of_sfdecl fdecl =
  string_of_typ fdecl.styp ^ " " ^
  fdecl.sname ^ "(" ^ String.concat ", " (List.map snd fdecl.sparams) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.slocals) ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_sfdecl funcs)
